﻿/*//////////////////////////////////////////////////////////////////////////////////////////////////////////

  FIVe3D v2.1.2  -  2008-10-12
  Flash Interactive Vector-based 3D
  Mathieu Badimon  |  five3d.mathieu-badimon.com  |  www.mathieu-badimon.com  |  contact@mathieu-badimon.com

  Modified by Devin Reimer (blog.almostlogical.com) | devin@almostlogical.com
  - modified to allow for JigLibFlash support (matrix only modifications)
/*//////////////////////////////////////////////////////////////////////////////////////////////////////////

package five3D.display {

	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.display.Graphics;
	import flash.display.Sprite;
	import flash.geom.ColorTransform;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import five3D.display.Bitmap3D;
	import five3D.display.Graphics3D;
	import five3D.display.Scene3D;
	import five3D.display.Shape3D;
	import five3D.display.Sprite2D;
	import five3D.geom.Matrix3D;
	import five3D.geom.Point3D;
	import five3D.utils.ColorUtilities;

	public class Sprite3D extends Sprite {

		private var __visible:Boolean = true;
		private var __x:Number = 0;
		private var __y:Number = 0;
		private var __z:Number = 0;
		private var __scalex:Number = 1;
		private var __scaley:Number = 1;
		private var __scalez:Number = 1;
		private var __rotationx:Number = 0;
		private var __rotationy:Number = 0;
		private var __rotationz:Number = 0;
		private var __matrix:Matrix3D;
		internal var __concatenatedmatrix:Matrix3D;
		private var __normal:Point3D;
		private var __graphics:Graphics3D;
		private var __singlesided:Boolean = false;
		private var __flatshaded:Boolean = false;
		private var __childrensorted:Boolean = false;
		private var __render:Boolean = true;
		private var __renderculling:Boolean = false;
		private var __culling:Boolean = false;
		private var __rendershading:Boolean = false;
		private var __flatshading:Boolean = false;
		private var __matrixOnly:Boolean = false; //the only way to perform rotations, move objects, get position or rotation directly from the matrix itself (implementation where this does not need to be locked down can be added in the future)

		public function Sprite3D() {
			__matrix = new Matrix3D();
			__graphics = new Graphics3D();
		}

		override public function get visible():Boolean {
			return __visible;
		}

		override public function set visible(value:Boolean):void {
			__visible = value;
		}

		override public function get mouseX():Number {
			var scene:Scene3D = getScene();
			if (scene == null) throw new Error("Cannot access this property without having the class connected to the Scene3D class through the display list hierarchy.");
			else {
				if (__concatenatedmatrix == null) throw new Error("Cannot access this property before the class has been rendered once.");
				else return __concatenatedmatrix.getInverseCoordinates(scene.mouseX, scene.mouseY, scene.viewDistance).x;
			}
		}

		override public function get mouseY():Number {
			var scene:Scene3D = getScene();
			if (scene == null) throw new Error("Cannot access this property without having the class connected to the Scene3D class through the display list hierarchy.");
			else {
				if (__concatenatedmatrix == null) throw new Error("Cannot access this property before the class has been rendered once.");
				else return __concatenatedmatrix.getInverseCoordinates(scene.mouseX, scene.mouseY, scene.viewDistance).y;
			}
		}

		public function get mouseXY():Point {
			var scene:Scene3D = getScene();
			if (scene == null) throw new Error("Cannot access this property without having the class connected to the Scene3D class through the display list hierarchy.");
			else {
				if (__concatenatedmatrix == null) throw new Error("Cannot access this property before the class has been rendered once.");
				else return __concatenatedmatrix.getInverseCoordinates(scene.mouseX, scene.mouseY, scene.viewDistance);
			}
		}

		private function getScene():Scene3D {
			for (var i:DisplayObjectContainer = DisplayObjectContainer(this); i; i = i.parent) if (i is Scene3D) return Scene3D(i);
			return null;
		}

		override public function get x():Number {
			
			if (!__matrixOnly) return __x;
			throw new Error("Can not get x value when in matrix only mode");
		}

		override public function set x(value:Number):void {
			if (!__matrixOnly)
			{
				__x = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set x value when in matrix only mode");
			}
		}
		
		override public function get y():Number {
			if (!__matrixOnly) return __y;
			throw new Error("Can not get y value when in matrix only mode");
		}
		

		override public function set y(value:Number):void {
			if (!__matrixOnly)
			{
				__y = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set y value when in matrix only mode");
			}
		}

		public function get z():Number {
			if (!__matrixOnly) return __z;
			throw new Error("Can not get z value when in matrix only mode");
		}

		public function set z(value:Number):void {
			if (!__matrixOnly)
			{
				__z = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set z value when in matrix only mode");
			}
		}

		override public function get scaleX():Number {
			if (!__matrixOnly) return __scalex;
			throw new Error("Can not get scaleX value when in matrix only mode");
		}

		override public function set scaleX(value:Number):void {
			if (!__matrixOnly)
			{
				__scalex = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set scaleX value when in matrix only mode");
			}
		}

		override public function get scaleY():Number {
			if (!__matrixOnly) return __scaley;
			throw new Error("Can not get scaleY value when in matrix only mode");
		}

		override public function set scaleY(value:Number):void {
			if (!__matrixOnly)
			{
				__scaley = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set scaleY value when in matrix only mode");
			}
		}

		public function get scaleZ():Number {
			if (!__matrixOnly) return __scalez;
			throw new Error("Can not get scaleZ value when in matrix only mode");
		}

		public function set scaleZ(value:Number):void {
			if (!__matrixOnly)
			{
				__scalez = value;
				askRendering();
			}
			else
			{
				throw new Error("Can not set scaleZ value when in matrix only mode");
			}
		}

		public function get rotationX():Number {
			if (!__matrixOnly) return __rotationx;
			throw new Error("Can not get rotationX value when in matrix only mode");
		}

		public function set rotationX(value:Number):void {
			if (!__matrixOnly)
			{
				__rotationx = formatRotation(value);
				askRendering();
			}
			else
			{
				throw new Error("Can not set rotationX value when in matrix only mode");
			}
		}

		public function get rotationY():Number {
			if (!__matrixOnly) return __rotationy;
			throw new Error("Can not get rotationY value when in matrix only mode");
		}

		public function set rotationY(value:Number):void {
			if (!__matrixOnly)
			{
				__rotationy = formatRotation(value);
				askRendering();
			}
			else
			{
				throw new Error("Can not set rotationY value when in matrix only mode");
			}
		}

		public function get rotationZ():Number {
			if (!__matrixOnly) return __rotationz;
			throw new Error("Can not get rotationZ value when in matrix only mode");
		}

		public function set rotationZ(value:Number):void {
			if (!__matrixOnly)
			{
				__rotationz = formatRotation(value);
				askRendering();
			}
			else
			{
				throw new Error("Can not set rotationZ value when in matrix only mode");
			}
		}

		static private function formatRotation(angle:Number):Number {
			var angle2:Number = angle%360;
			if (angle2 < -180) return angle2+360;
			if (angle2 > 180) return angle2-360;
			return angle2;
		}

		public function get graphics3D():Graphics3D {
			return __graphics;
		}

		public function get singleSided():Boolean {
			return __singlesided;
		}

		public function set singleSided(value:Boolean):void {
			__singlesided = value;
			if (__singlesided) __renderculling = true;
			else {
				__renderculling = false;
				__culling = false;
			}
		}

		public function get flatShaded():Boolean {
			return __flatshaded;
		}

		public function set flatShaded(value:Boolean):void {
			__flatshaded = value;
			if (__flatshaded) __rendershading = true;
			else {
				__rendershading = false;
				if (__flatshading) removeFlatShading();
			}
		}

		public function get childrenSorted():Boolean {
			return __childrensorted;
		}

		public function set childrenSorted(value:Boolean):void {
			__childrensorted = value;
		}

		internal function askRendering():void {
			__render = true;
			__graphics.askRendering();
			if (__singlesided) __renderculling = true;
			if (__flatshaded) __rendershading = true;
			for (var i:int = 0; i < numChildren; i++) getChildAtCasted(i).askRendering();
		}

		internal function askRenderingShading():void {
			if (__flatshaded) __rendershading = true;
			for (var i:int = 0; i < numChildren; i++) getChildAtCasted(i).askRenderingShading();
		}

		private function getChildAtCasted(index:int):* {
			if (getChildAt(index) is Sprite3D) return Sprite3D(getChildAt(index));
			if (getChildAt(index) is Shape3D) return Shape3D(getChildAt(index));
			if (getChildAt(index) is Bitmap3D) return Bitmap3D(getChildAt(index));
			if (getChildAt(index) is Sprite2D) return Sprite2D(getChildAt(index));
		}

		internal function render(scene:Scene3D):void {
			if (!__visible && super.visible) super.visible = false;
			else if (__visible) {
				if (__render) {
					getMatrices();
					__render = false;
				}
				if (__renderculling) {
					getCulling(scene.viewDistance);
					__renderculling = false;
				}
				if (__culling) {
					if (super.visible) super.visible = false;
				}
				else {
					if (!super.visible) super.visible = true;
					
					__graphics.render(super.graphics, __concatenatedmatrix, scene.viewDistance);
					if (__rendershading) {
						setFlatShading(scene);
						__rendershading = false;
					}
					for (var i:int = 0; i < numChildren; i++) getChildAtCasted(i).render(scene);
					if (__childrensorted) sortChildren();
				}
			}
		}

		private function getMatrices():void {
			if (!__matrixOnly) //if not in matrix only mode, create own matrix based on values (added Devin Reimer)
			{
				__matrix.createBox(__scalex, __scaley, __scalez, degreesToRadians(__rotationx), degreesToRadians(__rotationy), degreesToRadians(__rotationz), __x, __y, __z);
			}
			if (parent is Scene3D) {
				__concatenatedmatrix = __matrix.clone();
			}
			else {
				__concatenatedmatrix = Sprite3D(parent).__concatenatedmatrix.clone();
				__concatenatedmatrix.concat(__matrix);
			}
			__normal = null;
		}

		static private function degreesToRadians(angle:Number):Number {
			return angle*Math.PI/180;
		}

		private function getCulling(viewdistance:Number):void {
			var point1:Point3D = __concatenatedmatrix.transformPoint(new Point3D(0, 0, 0));
			var point2:Point3D = __concatenatedmatrix.transformPoint(new Point3D(0, 0, 1));
			__normal = point2.subtract(point1);
			var camera:Point3D = new Point3D(point1.x, point1.y, point1.z+viewdistance);
			__culling = __normal.dot(camera) < 0;
		}

		private function setFlatShading(scene:Scene3D):void {
			if (__normal == null) {
				var point1:Point3D = __concatenatedmatrix.transformPoint(new Point3D(0, 0, 0));
				var point2:Point3D = __concatenatedmatrix.transformPoint(new Point3D(0, 0, 1));
				__normal = point2.subtract(point1);
			}
			__normal.normalize(1);
			var dot:Number = __normal.dot(scene.ambientLightVectorNormalized);
			var brightness:Number = dot*scene.ambientLightIntensity;
			var colortransform:ColorTransform = new ColorTransform();
			ColorUtilities.setBrightness(colortransform, brightness);
			colortransform.alphaMultiplier = alpha;
			transform.colorTransform = colortransform;
			__flatshading = true;
		}

		private function removeFlatShading():void {
			transform.colorTransform = new ColorTransform();
			__flatshading = false;
		}

		private function sortChildren():void {
			var sort:Array = [];
			var child:*;
			for (var i:int = 0; i < numChildren; i++) {
				child = getChildAtCasted(i);
				if (child.visible) sort.push({child:child, z:child.__concatenatedmatrix.tz});
			}
			sort.sortOn("z", Array.DESCENDING | Array.NUMERIC);
			for (i = 0; i < sort.length; i++) setChildIndex(sort[i].child, i);
		}
		
		//following functions added by Devin Reimer (blog.almostlogical.com)
		
		public function get matrix():Matrix3D { return __matrix; }
		public function set matrix(val:Matrix3D):void { __matrix = val; askRendering(); }
		
		//once this is changed it should not be undone, as no positions or rotations will be set
		// once place this function is use is when you are using external libaries that modify this matrix directly. ex: JigLibFlash
		public function onlyAllowMatrixModifications():void { __matrixOnly = true; }
		//removes matrix only modifications, this position and rotation will revert to the last placement before onlyAllowMatrixModifications was enabled
		public function reset():void { __matrixOnly = false; }
		
		//end of functions added by Devin Reimer (blog.almostlogical.com)

		// Errors

		override public function get height():Number {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function set height(value:Number):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get rotation():Number {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function set rotation(value:Number):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get scale9Grid():Rectangle {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function set scale9Grid(value:Rectangle):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get scrollRect():Rectangle {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function set scrollRect(value:Rectangle):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get width():Number {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function set width(value:Number):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function globalToLocal(point:Point):Point {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function localToGlobal(point:Point):Point {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get dropTarget():DisplayObject {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function get graphics():Graphics {
			throw new Error("The Sprite3D class does not implement this property or method (use \"graphics3D\" instead).");
		}

		override public function startDrag(lockCenter:Boolean = false, bounds:Rectangle = null):void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}

		override public function stopDrag():void {
			throw new Error("The Sprite3D class does not implement this property or method.");
		}
	}

}